/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { CollaborationListResource } from '@alfa-client/collaboration-shared';
import { SnackBarService } from '@alfa-client/ui';
import { TestBed } from '@angular/core/testing';
import { CommandLinkRel, CommandResource } from '@alfa-client/command-shared';
import { StateResource, createEmptyStateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { UntypedFormBuilder } from '@angular/forms';
import { CollaborationService } from 'libs/collaboration-shared/src/lib/collaboration.service';
import { createCollaborationListResource } from 'libs/collaboration-shared/test/collaboration';
import { createCommandErrorResource, createCommandResource } from 'libs/command-shared/test/command';
import { singleColdCompleted } from 'libs/tech-shared/test/marbles';
import { Observable, of } from 'rxjs';
import { CollaborationRequestFormService } from './collaboration.request.formservice';

describe('CollaborationRequestFormService', () => {
  let formService: CollaborationRequestFormService;

  let service: Mock<CollaborationService>;
  let snackBarService: Mock<SnackBarService>;

  const collaborationListResource: CollaborationListResource = createCollaborationListResource();

  beforeEach(() => {
    service = mock(CollaborationService);
    snackBarService = mock(SnackBarService);

    TestBed.configureTestingModule({
      providers: [
        CollaborationRequestFormService,
        UntypedFormBuilder,
        { provide: CollaborationService, useValue: service },
        { provide: SnackBarService, useValue: snackBarService },
      ],
    });

    formService = TestBed.inject(CollaborationRequestFormService);
    TestBed.inject(CollaborationRequestFormService);
  });

  it('should create', () => {
    expect(formService).toBeTruthy();
  });

  describe('form', () => {
    it('should have initial zustaendigeStelle value', () => {
      expect(formService.form.controls[CollaborationRequestFormService.FIELD_ZUSTAENDIGE_STELLE].value).toBeNull();
    });

    it('should have initial titel value', () => {
      expect(formService.form.controls[CollaborationRequestFormService.FIELD_TITEL].value).toBeNull();
    });

    it('should have initial beschreibung value', () => {
      expect(formService.form.controls[CollaborationRequestFormService.FIELD_BESCHREIBUNG].value).toBeNull();
    });
  });

  describe('do submit', () => {
    const stateCommandResource: StateResource<CommandResource> = createStateResource(
      createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]),
    );

    beforeEach(() => {
      formService.listResource = collaborationListResource;
      service.create.mockReturnValue(of(stateCommandResource));
    });

    it('should call service', () => {
      formService.submit();

      expect(service.create).toHaveBeenCalledWith(collaborationListResource, formService.form.value);
    });

    it('should call service refresh', () => {
      formService.submit().subscribe();

      expect(service.refreshList).toHaveBeenCalled();
    });

    it('should call handleCommandError', () => {
      const handleCommanErrorSpy = jest.spyOn(formService as any, 'handleCommandError');

      formService.submit().subscribe();

      expect(handleCommanErrorSpy).toHaveBeenCalledWith(stateCommandResource.resource);
    });

    it('should return stateCommandResource', () => {
      const response$: Observable<StateResource<CommandResource>> = formService.submit();

      expect(response$).toBeObservable(singleColdCompleted(stateCommandResource));
    });
  });

  describe('handleCommandError', () => {
    let showErrorSpy: jest.SpyInstance;

    beforeEach(() => {
      showErrorSpy = jest.spyOn(formService as any, 'showError');
    });

    it('should not do anything if commandResource is nil', () => {
      service.create.mockReturnValue(of(createEmptyStateResource()));

      formService.submit().subscribe();

      expect(showErrorSpy).not.toHaveBeenCalled();
    });

    it('should call showError if commandResource has error', () => {
      service.create.mockReturnValue(of(createStateResource(createCommandErrorResource())));

      formService.submit().subscribe();

      expect(showErrorSpy).toHaveBeenCalled();
    });
  });

  describe('showError', () => {
    it('should call snackBarService.showError with commandResource errorMessage', () => {
      const commandResourceWithError: CommandResource = createCommandErrorResource();
      service.create.mockReturnValue(of(createStateResource(commandResourceWithError)));

      formService.submit().subscribe();

      expect(snackBarService.showError).toHaveBeenCalledWith(CollaborationRequestFormService.GENERIC_ERROR_MESSAGE);
    });
  });

  describe('set list resource', () => {
    it('should set given list resource', () => {
      formService.listResource = undefined;

      formService.setListResource(collaborationListResource);

      expect(formService.listResource).toBe(collaborationListResource);
    });
  });
});
