/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { fakeAsync, tick } from '@angular/core/testing';
import { faker } from '@faker-js/faker';
import { Resource, ResourceUri } from '@ngxp/rest';
import { cold } from 'jest-marbles';
import { DummyLinkRel, DummyListLinkRel } from 'libs/tech-shared/test/dummy';
import { createDummyListResource, createDummyResource, createFilledDummyListResource } from 'libs/tech-shared/test/resource';
import { BehaviorSubject, Observable, of } from 'rxjs';
import { singleCold, singleHot } from '../../../test/marbles';
import { EMPTY_ARRAY } from '../tech.util';
import { ResourceListService } from './list-resource.service';
import { CreateResourceData, LinkRelationName, ListItemResource, ListResourceServiceConfig } from './resource.model';
import { ResourceRepository } from './resource.repository';
import { ListResource, StateResource, createEmptyStateResource, createStateResource } from './resource.util';

import * as ResourceUtil from './resource.util';

describe('ListResourceService', () => {
  let service: ResourceListService<Resource, ListResource, ListItemResource>;
  let config: ListResourceServiceConfig<Resource>;
  let resourceRepository: Mock<ResourceRepository>;

  const listLinkRel: LinkRelationName = DummyListLinkRel.LIST;
  const createLinkRel: LinkRelationName = DummyLinkRel.DUMMY;
  const listResourceListLinkRel: LinkRelationName = DummyListLinkRel.LIST;
  const listResource: ListResource = createDummyListResource([listLinkRel, createLinkRel]);

  const baseResource: Resource = createDummyResource();
  const baseStateResource: StateResource<Resource> = createStateResource(baseResource);
  const baseResourceSubj: BehaviorSubject<StateResource<Resource>> = new BehaviorSubject<StateResource<Resource>>(
    baseStateResource,
  );

  const linkedResourceUri = faker.internet.url();

  beforeEach(() => {
    config = {
      baseResource: baseResourceSubj,
      listLinkRel,
      listResourceListLinkRel,
      createLinkRel,
    };
    resourceRepository = mock(ResourceRepository);

    service = new ResourceListService(config, useFromMock(resourceRepository));
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  describe('getList', () => {
    const listStateResource: StateResource<ListResource> = createStateResource(createDummyListResource());

    let isInvalidResourceCombinationSpy: jest.SpyInstance;

    beforeEach(() => {
      service.listResource.next(listStateResource);

      service.handleNullConfigResource = jest.fn();
      service.handleChanges = jest.fn();
      isInvalidResourceCombinationSpy = jest.spyOn(ResourceUtil, 'isInvalidResourceCombination').mockReturnValue(true);
    });

    it('should handle config resource changed', fakeAsync(() => {
      service.getList().subscribe();
      tick();

      expect(service.handleChanges).toHaveBeenCalledWith(listStateResource, baseResource);
    }));

    it('should handle null configresource', fakeAsync(() => {
      service.getList().subscribe();
      tick();

      expect(service.handleNullConfigResource).toHaveBeenCalledWith(baseResource);
    }));

    it('should call isInvalidResourceCombinationSpy', fakeAsync(() => {
      service.getList().subscribe();
      tick();

      expect(isInvalidResourceCombinationSpy).toHaveBeenCalled();
    }));

    it('should return initial value', () => {
      service.listResource.asObservable = jest.fn().mockReturnValue(singleHot(listResource, '-a'));

      const apiRootStateResource$: Observable<StateResource<Resource>> = service.getList();

      expect(apiRootStateResource$).toBeObservable(cold('a', { a: createEmptyStateResource(true) }));
    });
  });

  describe('handle changes', () => {
    const listStateResource: StateResource<ListResource> = createStateResource(createDummyListResource());
    const changedConfigResource: Resource = createDummyResource();

    describe('on different config resource', () => {
      it('should handle changes', () => {
        service.handleConfigResourceChanges = jest.fn();
        service.baseResource = createDummyResource();

        service.handleChanges(listStateResource, changedConfigResource);

        expect(service.handleConfigResourceChanges).toHaveBeenCalledWith(changedConfigResource);
      });
    });

    describe('on same config resource', () => {
      const listStateResource: StateResource<ListResource> = createStateResource(createDummyListResource());

      beforeEach(() => {
        service.baseResource = baseResource;
      });

      it('should call shouldLoadResource', () => {
        service.shouldLoadResource = jest.fn();

        service.handleChanges(listStateResource, baseResource);

        expect(service.shouldLoadResource).toHaveBeenCalledWith(listStateResource, baseResource);
      });

      it('should load resource', () => {
        service.shouldLoadResource = jest.fn().mockReturnValue(true);
        service.loadListResource = jest.fn();

        service.handleChanges(listStateResource, baseResource);

        expect(service.loadListResource).toHaveBeenCalledWith(baseResource, listLinkRel);
      });

      it('should NOT load resource on shouldLoadResource false', () => {
        service.loadListResource = jest.fn();
        service.shouldLoadResource = jest.fn().mockReturnValue(false);

        service.handleChanges(listStateResource, baseResource);

        expect(service.loadListResource).not.toHaveBeenCalled();
      });
    });
  });

  describe('handle config resource changes', () => {
    const changedConfigResource: Resource = createDummyResource();

    it('should update base resource', () => {
      service.baseResource = null;

      service.handleConfigResourceChanges(changedConfigResource);

      expect(service.baseResource).toBe(changedConfigResource);
    });

    describe('on existing list link rel', () => {
      it('should load list on stable state resource', () => {
        service.loadListResource = jest.fn();
        service.listResource.next(createStateResource(createDummyListResource()));
        const configResuorce: Resource = createDummyListResource([listLinkRel]);

        service.handleConfigResourceChanges(configResuorce);

        expect(service.loadListResource).toHaveBeenCalledWith(service.baseResource, listLinkRel);
      });

      it('should NOT load list on stable unstable resource', () => {
        jest.spyOn(ResourceUtil, 'isStateResoureStable').mockReturnValue(false);
        service.loadListResource = jest.fn();
        service.listResource.next(createStateResource(createDummyListResource()));
        const configResuorce: Resource = createDummyListResource([listLinkRel]);

        service.handleConfigResourceChanges(configResuorce);

        expect(service.loadListResource).not.toHaveBeenCalled();
      });
    });

    describe('on missing list link rel', () => {
      it('should clear current list resource on stable state resource', () => {
        jest.spyOn(ResourceUtil, 'isStateResoureStable').mockReturnValue(true);
        service.listResource.next(createStateResource(createDummyListResource()));

        const configResuorceWithoutLink: Resource = createDummyListResource();

        service.handleConfigResourceChanges(configResuorceWithoutLink);

        expect(service.listResource.value).toEqual(createEmptyStateResource());
      });

      it('should keep current current list resource on unstable state resource', () => {
        jest.spyOn(ResourceUtil, 'isStateResoureStable').mockReturnValue(false);
        const currentListStateResource: StateResource<ListResource> = createStateResource(createDummyListResource());
        service.listResource.next(currentListStateResource);

        const configResuorceWithoutLink: Resource = createDummyListResource();

        service.handleConfigResourceChanges(configResuorceWithoutLink);

        expect(service.listResource.value).toBe(currentListStateResource);
      });
    });
  });

  describe('load list resource', () => {
    const linkRel: string = faker.person.firstName();
    const loadedListResource: ListResource = createDummyListResource();
    const resource: Resource = createDummyResource();

    beforeEach(() => {
      service.setStateResourceLoading = jest.fn();
      resourceRepository.getListResource.mockReturnValue(of(loadedListResource));
    });

    it('should set state resource to loading', () => {
      service.loadListResource(resource, linkRel);

      expect(service.setStateResourceLoading).toHaveBeenCalled();
    });

    it('should call repository', () => {
      service.loadListResource(resource, linkRel);

      expect(resourceRepository.getListResource).toHaveBeenCalledWith(resource, linkRel);
    });

    it('should update stateresource', () => {
      service.updateListResource = jest.fn();

      service.loadListResource(resource, linkRel);

      expect(service.updateListResource).toHaveBeenCalledWith(loadedListResource);
    });
  });

  describe('getLinkedResource', () => {
    beforeEach(() => {
      service.linkedResources[linkedResourceUri] = new BehaviorSubject(createEmptyStateResource());
    });

    it('should call createIfNotExist ', () => {
      service.createIfNotExist = jest.fn();
      service.loadLinkedResoure = jest.fn();

      service.getLinkedResource(linkedResourceUri);

      expect(service.createIfNotExist).toHaveBeenCalled();
    });
  });

  describe('createIfNotExists', () => {
    it('should create linked resource if none exists', () => {
      service.createIfNotExist(linkedResourceUri);

      expect(service.linkedResources[linkedResourceUri].value).toEqual(createEmptyStateResource());
    });
  });

  describe('loadLinkedResourceIfRequired', () => {
    it('should set Loading in Linked Resource', () => {
      service.createIfNotExist(linkedResourceUri);
      service.setLinkedResourceLoading = jest.fn();
      service.loadLinkedResoure = jest.fn();

      service.getLinkedResource(linkedResourceUri);

      expect(service.setLinkedResourceLoading).toHaveBeenCalledWith(linkedResourceUri);
    });

    it('should call loadLinkedResource if loading is required', () => {
      service.createIfNotExist(linkedResourceUri);
      service.loadLinkedResoure = jest.fn();

      service.getLinkedResource(linkedResourceUri);

      expect(service.loadLinkedResoure).toHaveBeenCalledWith(linkedResourceUri);
    });
  });

  describe('loadLinkedResource', () => {
    it('should call repository getResource', () => {
      resourceRepository.getResource = jest.fn().mockReturnValue(of(createDummyResource()));
      service.createIfNotExist(linkedResourceUri);

      service.loadLinkedResoure(linkedResourceUri);

      expect(resourceRepository.getResource).toHaveBeenCalledWith(linkedResourceUri);
    });
  });

  describe('create', () => {
    const toCreate: unknown = {};

    beforeEach(() => {
      service.listResource.value.resource = listResource;
    });

    it('should throw error if link is not present', () => {
      service.listResource.next(createStateResource(createDummyListResource()));

      expect(() => service.create(toCreate)).toThrowError('No creation link exists.');
    });

    it('should throw error if list resource is empty', () => {
      service.listResource.next(createEmptyStateResource());

      expect(() => service.create(toCreate)).toThrowError('No list resource available.');
    });

    it('should trigger resourceLoader', () => {
      service.create(toCreate);

      const createResourceData: CreateResourceData<ListResource> = {
        resource: listResource,
        toCreate,
        linkRel: createLinkRel,
      };
      expect(resourceRepository.createResource).toHaveBeenCalledWith(createResourceData);
    });

    it('should return value', () => {
      const returnResource: Resource = createDummyResource();
      resourceRepository.createResource.mockReturnValue(singleHot(returnResource));

      const createdResource: Observable<Resource> = service.create(toCreate);

      expect(createdResource).toBeObservable(singleCold(returnResource));
    });
  });

  describe('select', () => {
    const selfHref: ResourceUri = 'dummySelfHref';
    const dummyResource: Resource = createResourceWithUri(selfHref);

    const listResource: ListResource = createFilledDummyListResource([dummyResource]);
    const listStateResource: StateResource<ListResource> = createStateResource(listResource);

    const loadedResource: Resource = createDummyResource();

    function createResourceWithUri(href: ResourceUri) {
      const _links = {
        self: { href },
      };
      return {
        ...createDummyResource(),
        _links,
      };
    }

    beforeEach(() => {
      service.listResource.next(listStateResource);
      resourceRepository.getResource.mockReturnValue(of(loadedResource));
    });

    it('should set resource loading', () => {
      service.setSelectedResourceLoading = jest.fn();

      service.select(selfHref);

      expect(service.setSelectedResourceLoading).toHaveBeenCalled();
    });

    it('should call repository with uri', fakeAsync(() => {
      service.select(selfHref);
      tick();

      expect(resourceRepository.getResource).toHaveBeenCalledWith(selfHref);
    }));

    it('should set loaded resource as selected', fakeAsync(() => {
      service.select(selfHref);
      tick();

      expect(service.selectedResource.value).toEqual(createStateResource(loadedResource));
    }));
  });

  describe('setSelectedResourceLoading', () => {
    it('should set loading on selected resource', () => {
      service.selectedResource.next(createEmptyStateResource());

      service.setSelectedResourceLoading();

      expect(service.selectedResource.value.loading).toBeTruthy();
    });
  });

  describe('unselect', () => {
    it('should clear current selected resource', () => {
      service.selectedResource.next(createStateResource(createDummyResource()));

      service.unselect();

      expect(service.selectedResource.value).toEqual(createEmptyStateResource());
    });
  });

  describe('getSelected', () => {
    it('should return selected resource', (done) => {
      const dummyStateResource: StateResource<Resource> = createStateResource(createDummyResource());

      service.getSelected().subscribe((selected) => {
        expect(selected).toEqual(dummyStateResource);
        done();
      });

      service.selectedResource.next(dummyStateResource);
    });
  });

  describe('refresh', () => {
    it('should trigger reload current listResource', () => {
      service.refresh();

      expect(service.listResource.value.reload).toEqual(true);
    });
  });

  describe('prev', () => {
    const previousListResource: ListResource = createDummyListResource();

    it('should throw error if prev link is not present', () => {
      service.listResource.next(createStateResource(listResource));

      expect(() => service.prev()).toThrowError('There is no previous page.');
    });

    it('should call resourceLoader with prev link', () => {
      resourceRepository.getListResource.mockReturnValue(of(previousListResource));
      const listResource: ListResource = createDummyListResource([service.prevLink]);
      service.listResource.next(createStateResource(listResource));

      service.prev();

      expect(resourceRepository.getListResource).toHaveBeenCalledWith(listResource, service.prevLink);
    });

    it('should update listResource', () => {
      const listResource: ListResource = createDummyListResource([service.prevLink]);
      service.listResource.next(createStateResource(listResource));
      resourceRepository.getListResource.mockReturnValue(of(previousListResource));

      service.prev();

      expect(service.listResource.value.resource).toBe(previousListResource);
    });
  });

  describe('next', () => {
    it('should throw error if next link is not present', () => {
      service.listResource.next(createStateResource(listResource));

      expect(() => service.next()).toThrowError('There is no next page.');
    });

    it('should call resourceLoader with next link', () => {
      const nextListResource: ListResource = createDummyListResource();
      resourceRepository.getListResource.mockReturnValue(of(nextListResource));
      const listResourceWithPrevLink: ListResource = createDummyListResource([service.nextLink]);
      service.listResource.next(createStateResource(listResourceWithPrevLink));

      service.next();

      expect(resourceRepository.getListResource).toHaveBeenCalledWith(listResourceWithPrevLink, service.nextLink);
    });

    it('should update listResource', () => {
      const listResourceWithNextLink: ListResource = createDummyListResource([service.nextLink]);
      service.listResource.next(createStateResource(listResourceWithNextLink));
      const nextListResource: ListResource = createDummyListResource();
      resourceRepository.getListResource.mockReturnValue(of(nextListResource));

      service.next();

      expect(service.listResource.value.resource).toBe(nextListResource);
    });
  });

  describe('hasMore', () => {
    it('should return true if listResource has next link', () => {
      const listResourceWithLink: ListResource = createDummyListResource([service.nextLink]);
      service.listResource.next(createStateResource(listResourceWithLink));

      const hasMore: boolean = service.hasMore();

      expect(hasMore).toBe(true);
    });
  });

  describe('isFirst', () => {
    it('should return true if listResource has no prev link', () => {
      const listResource: ListResource = createDummyListResource();
      service.listResource.next(createStateResource(listResource));

      const isFirst: boolean = service.isFirst();

      expect(isFirst).toBe(true);
    });
  });

  describe('get items', () => {
    const listResourceItems: ListItemResource[] = [createDummyResource()];
    const stateListResource: StateResource<ListResource> = createStateResource(createFilledDummyListResource(listResourceItems));
    let getListSpy: jest.SpyInstance;

    beforeEach(() => {
      getListSpy = service.getList = jest.fn().mockReturnValue(of(stateListResource));
    });

    it('should get list', () => {
      service.getItems();

      expect(getListSpy).toHaveBeenCalled();
    });

    it('should call get embedded resources', () => {
      const spy = jest.spyOn(ResourceUtil, 'getEmbeddedResources');
      service.getItems().subscribe();

      expect(spy).toHaveBeenCalledWith(stateListResource, listResourceListLinkRel);
    });

    it('should return empty array on empty stateResource', (done) => {
      service.getList = jest.fn().mockReturnValue(of(createEmptyStateResource()));

      service.getItems().subscribe((items) => {
        expect(items).toEqual(EMPTY_ARRAY);
        done();
      });
    });

    it('should return listresource items', (done) => {
      service.getItems().subscribe((items) => {
        expect(items).toEqual(listResourceItems);
        done();
      });
    });

    it('should not emit on loading state resource', () => {
      service.getList = jest.fn().mockReturnValue(of(createEmptyStateResource(true)));

      const spy = jest.spyOn(ResourceUtil, 'getEmbeddedResources').mockClear();
      service.getItems().subscribe();

      expect(spy).not.toHaveBeenCalled();
    });
  });
});
