/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import {
  Bescheid,
  BescheidLinkRel,
  BescheidResource,
  BescheidSendBy,
  BescheidService,
  DocumentResource,
} from '@alfa-client/bescheid-shared';
import { BinaryFileResource } from '@alfa-client/binary-file-shared';
import { CommandResource, tapOnCommandSuccessfullyDone } from '@alfa-client/command-shared';
import {
  AbstractFormService,
  EMPTY_STRING,
  StateResource,
  convertToBoolean,
  formatForDatabase,
  isLoaded,
  isNotEmpty,
  isNotNil,
} from '@alfa-client/tech-shared';
import { VorgangWithEingangResource } from '@alfa-client/vorgang-shared';
import { Injectable, OnDestroy } from '@angular/core';
import {
  FormControl,
  UntypedFormArray,
  UntypedFormBuilder,
  UntypedFormControl,
  UntypedFormGroup,
} from '@angular/forms';
import { Resource, ResourceUri, getUrl, hasLink } from '@ngxp/rest';
import { isEmpty, isNil, isUndefined } from 'lodash-es';
import {
  BehaviorSubject,
  Observable,
  Subject,
  Subscription,
  combineLatest,
  filter,
  first,
  map,
  startWith,
} from 'rxjs';

@Injectable()
export class BescheidenFormService extends AbstractFormService implements OnDestroy {
  static readonly FIELD_BESCHIEDEN_AM = 'beschiedenAm';
  static readonly FIELD_BEWILLIGT = 'bewilligt';
  static readonly FIELD_BESCHEID_DOCUMENT = 'bescheidDocument';
  static readonly FIELD_ATTACHMENTS = 'attachments';
  public static readonly FIELD_SEND_BY = 'sendBy';
  static readonly FIELD_NACHRICHT_SUBJECT = 'nachrichtSubject';
  static readonly FIELD_NACHRICHT_TEXT = 'nachrichtText';

  static readonly FIELD_PATH_PREFIX = 'command.body';

  static readonly BETREFF_DEFAULT: string = 'Ihr Bescheid zum Antrag';

  private readonly bescheidChanges$: BehaviorSubject<Bescheid>;
  private readonly fileDelete$: Subject<BinaryFileResource>;
  private readonly showMissingBescheidDocumentError$: BehaviorSubject<boolean> =
    new BehaviorSubject<boolean>(false);

  private readonly activeStep$: BehaviorSubject<number> = new BehaviorSubject(1);
  readonly sendByManual: BehaviorSubject<boolean> = new BehaviorSubject(false);

  vorgangWithEingangResource: VorgangWithEingangResource;

  private formControlSubscriptions: Subscription;
  private formChangesSubscription: Subscription;

  constructor(
    formBuilder: UntypedFormBuilder,
    private readonly bescheidService: BescheidService,
  ) {
    super(formBuilder);

    this.bescheidChanges$ = new BehaviorSubject<Bescheid>(this.getFormValue());
    this.fileDelete$ = new Subject<BinaryFileResource>();
    this.init();
  }

  init(): void {
    this.formControlSubscriptions = this.subscribeToSendBy();
    this.initializeFormChanges();
  }

  subscribeToSendBy(): Subscription {
    return this.form.controls[BescheidenFormService.FIELD_SEND_BY].valueChanges.subscribe(
      (sendBy: BescheidSendBy) => this.updateSendByManual(sendBy),
    );
  }

  updateSendByManual(sendBy: BescheidSendBy): void {
    const isSendByManual = sendBy === BescheidSendBy.MANUAL;
    if (isSendByManual !== this.sendByManual.value) this.sendByManual.next(isSendByManual);
  }

  initializeFormChanges(): void {
    this.formChangesSubscription = this.form.valueChanges
      .pipe(
        startWith(this.getFormValue()),
        map((value) => this.getValueWithApproved(value)),
      )
      .subscribe((bescheid) => this.bescheidChanges$.next(bescheid));
  }

  ngOnDestroy(): void {
    this.unsubscribe();
  }

  private unsubscribe(): void {
    if (isNotNil(this.formChangesSubscription) && !this.formChangesSubscription.closed) {
      this.formChangesSubscription.unsubscribe();
    }
    if (isNotNil(this.formControlSubscriptions)) this.formControlSubscriptions.unsubscribe();
  }

  public patchValues(bescheid: BescheidResource): void {
    if (isNil(bescheid)) {
      return;
    }
    const bescheidDocumentUri: ResourceUri = this.getBescheidDocumentUri(bescheid);
    this.bescheidService.setDocumentUri(bescheidDocumentUri);
    this.patch({
      [BescheidenFormService.FIELD_BESCHIEDEN_AM]: bescheid.beschiedenAm,
      [BescheidenFormService.FIELD_BEWILLIGT]: String(bescheid.bewilligt),
      [BescheidenFormService.FIELD_BESCHEID_DOCUMENT]: bescheidDocumentUri,
      [BescheidenFormService.FIELD_SEND_BY]:
        isUndefined(bescheid.sendBy) ? BescheidSendBy.NACHRICHT : bescheid.sendBy,
      [BescheidenFormService.FIELD_NACHRICHT_SUBJECT]: bescheid.nachrichtSubject,
      [BescheidenFormService.FIELD_NACHRICHT_TEXT]: bescheid.nachrichtText,
    });
    bescheid.attachments.forEach((attachmentLink) =>
      (this.form.controls[BescheidenFormService.FIELD_ATTACHMENTS] as UntypedFormArray).push(
        new UntypedFormControl(attachmentLink),
      ),
    );
  }

  private getBescheidDocumentUri(bescheid: BescheidResource): ResourceUri {
    if (hasLink(bescheid, BescheidLinkRel.BESCHEID_DOCUMENT)) {
      return getUrl(bescheid, BescheidLinkRel.BESCHEID_DOCUMENT);
    }
    return null;
  }

  protected initForm(): UntypedFormGroup {
    return this.formBuilder.group({
      [BescheidenFormService.FIELD_BESCHIEDEN_AM]: new UntypedFormControl(new Date(Date.now())),
      [BescheidenFormService.FIELD_BEWILLIGT]: new UntypedFormControl('true'),
      [BescheidenFormService.FIELD_SEND_BY]: new UntypedFormControl(BescheidSendBy.NACHRICHT),
      [BescheidenFormService.FIELD_BESCHEID_DOCUMENT]: new UntypedFormControl(null),
      [BescheidenFormService.FIELD_ATTACHMENTS]: new UntypedFormArray([]),
      [BescheidenFormService.FIELD_NACHRICHT_SUBJECT]: new UntypedFormControl(''),
      [BescheidenFormService.FIELD_NACHRICHT_TEXT]: new UntypedFormControl(''),
    });
  }

  protected getPathPrefix(): string {
    return BescheidenFormService.FIELD_PATH_PREFIX;
  }

  protected doSubmit(): Observable<StateResource<Resource>> {
    if (this.isPatch() || this.bescheidService.existsBescheidDraft()) {
      return this.bescheidService.updateBescheid(this.getValue());
    } else {
      return this.bescheidService
        .createBescheid(this.vorgangWithEingangResource, this.getValue())
        .pipe(this.updateNachrichtOnSuccess());
    }
  }

  updateNachrichtOnSuccess(): (
    source: Observable<StateResource<CommandResource>>,
  ) => Observable<StateResource<CommandResource>> {
    return tapOnCommandSuccessfullyDone(() => {
      this.bescheidService
        .getBescheidDraft()
        .pipe(filter(isLoaded), first())
        .subscribe((bescheidStateResource: StateResource<BescheidResource>) => {
          this.setNachrichtSubject(bescheidStateResource.resource.nachrichtSubject);
          this.setNachrichtText(bescheidStateResource.resource.nachrichtText);
        });
    });
  }

  public submitDraft(): Observable<StateResource<Resource>> {
    return this.submit().pipe(
      tapOnCommandSuccessfullyDone(() => this.bescheidService.reloadCurrentVorgang()),
    );
  }

  public getValue(): Bescheid {
    const value: any = this.getFormValue();
    return {
      ...value,
      beschiedenAm: formatForDatabase(value.beschiedenAm),
      bewilligt: convertToBoolean(value.bewilligt),
      attachments: value.attachments,
    };
  }

  public getValueWithApproved(value: any): Bescheid {
    return {
      ...value,
      bewilligt: isNotNil(value.bewilligt) ? convertToBoolean(value.bewilligt) : false,
    };
  }

  public getBescheidChanges(): Observable<Bescheid> {
    return this.bescheidChanges$.asObservable();
  }

  public setVorgangWithEingangResource(
    vorgangWithEingangResource: VorgangWithEingangResource,
  ): void {
    this.vorgangWithEingangResource = vorgangWithEingangResource;
  }

  public getVorgangWithEingangResource(): VorgangWithEingangResource {
    return this.vorgangWithEingangResource;
  }

  public deleteFile(binaryFileResource: BinaryFileResource): void {
    this.fileDelete$.next(binaryFileResource);
  }

  public getFileDelete(): Observable<BinaryFileResource> {
    return this.fileDelete$.asObservable();
  }

  public setActiveStep(step: number) {
    this.activeStep$.next(step);
  }

  public getActiveStep(): Observable<number> {
    return this.activeStep$.asObservable();
  }

  public clearBescheidDocumentFile(): void {
    this.updateBescheidDocumentFile(null);
  }

  public updateBescheidDocumentFile(uri: ResourceUri): void {
    this.getBescheidDocumentControl().patchValue(uri);
  }

  private getBescheidDocumentControl(): FormControl {
    return <FormControl>this.form.controls[BescheidenFormService.FIELD_BESCHEID_DOCUMENT];
  }

  public isSendByManual(): Observable<boolean> {
    return this.sendByManual.asObservable();
  }

  public patchNachricht(documentResource: DocumentResource): void {
    const value = this.getValue();
    if (isEmpty(value.nachrichtSubject)) {
      if (isNotEmpty(documentResource.nachrichtSubject)) {
        this.setNachrichtSubject(documentResource.nachrichtSubject);
      }
    }

    if (isEmpty(value.nachrichtText)) {
      if (isNotEmpty(documentResource.nachrichtText)) {
        this.setNachrichtText(documentResource.nachrichtText);
      }
    }
  }

  public clearNachricht(): void {
    this.setNachrichtSubject(EMPTY_STRING);
    this.setNachrichtText(EMPTY_STRING);
  }

  private setNachrichtSubject(value: string): void {
    this.form.controls[BescheidenFormService.FIELD_NACHRICHT_SUBJECT].patchValue(value);
  }

  private setNachrichtText(value: string): void {
    this.form.controls[BescheidenFormService.FIELD_NACHRICHT_TEXT].patchValue(value);
  }

  public setSendBy(sendBy: BescheidSendBy): void {
    this.form.controls[BescheidenFormService.FIELD_SEND_BY].patchValue(sendBy);
  }

  public getShowMissingBescheidDocumentError(): Observable<boolean> {
    return combineLatest([
      this.showMissingBescheidDocumentError$.asObservable(),
      this.getBescheidChanges(),
    ]).pipe(
      map(
        ([showMissingBescheidDocumentError, bescheid]) =>
          showMissingBescheidDocumentError && isEmpty(bescheid.bescheidDocument),
      ),
    );
  }

  public validateBescheidDocumentExists(): boolean {
    this.showMissingBescheidDocumentError$.next(isEmpty(this.getValue().bescheidDocument));
    return isNotEmpty(this.getValue().bescheidDocument);
  }

  public isBetreffInvalid(): boolean {
    return this.form.controls[BescheidenFormService.FIELD_NACHRICHT_SUBJECT].invalid;
  }

  public isNachrichtInvalid(): boolean {
    return this.form.controls[BescheidenFormService.FIELD_NACHRICHT_TEXT].invalid;
  }
}
