/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.binaryfile;

import java.util.Map;
import java.util.Optional;

import jakarta.validation.ConstraintValidator;
import jakarta.validation.ConstraintValidatorContext;

import org.hibernate.validator.constraintvalidation.HibernateConstraintValidatorContext;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.util.unit.DataSize;
import org.springframework.util.unit.DataUnit;

class UploadBinaryFileSizeValidator implements ConstraintValidator<BinaryFileMaxSizeConstraint, UploadBinaryFileRequest> {

	static final String DEFAULT_UNIT_STRING = "MB";
	static final DataUnit DEFAULT_UNIT = DataUnit.fromSuffix(DEFAULT_UNIT_STRING);

	static final int DEFAULT_MAX_SIZE = 40;

	private final DataSize defaultDataSize = DataSize.of(DEFAULT_MAX_SIZE, DEFAULT_UNIT);

	@Autowired
	private BinaryFileProperties binaryFileProperties;

	@Override
	public boolean isValid(UploadBinaryFileRequest uploadRequest, ConstraintValidatorContext context) {
		var maxValidSize = getMaxValidSize(uploadRequest.getField());

		setDynamicPayload(buildDynamicPayload(maxValidSize), context);

		return isFileSizeValid(DataSize.ofBytes(uploadRequest.getSize()), maxValidSize);
	}

	private boolean isFileSizeValid(DataSize uploadFileSize, DataSize maxFileSize) {
		return uploadFileSize.toBytes() <= maxFileSize.toBytes();
	}

	DataSize getMaxValidSize(String field) {
		return Optional.ofNullable(binaryFileProperties.getMaxFileSize().get(field)).orElseGet(() -> defaultDataSize);
	}

	private void setDynamicPayload(DynamicViolationParameter dynamicViolationParameter, ConstraintValidatorContext context) {
		var unwrappedContext = getUnwrappedContext(HibernateConstraintValidatorContext.class, context);
		unwrappedContext.withDynamicPayload(dynamicViolationParameter);
	}

	<T> T getUnwrappedContext(Class<T> unwrapTo, ConstraintValidatorContext context) {
		return context.unwrap(unwrapTo);
	}

	private DynamicViolationParameter buildDynamicPayload(DataSize maxFileSize) {
		return DynamicViolationParameter.builder().map(Map.of("max", maxFileSize.toMegabytes())).build();
	}
}