/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.collaboration;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentMatcher;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.zufi.grpc.fachstelle.FachstelleServiceGrpc.FachstelleServiceBlockingStub;
import de.ozgcloud.zufi.grpc.fachstelle.GrpcFachstelleGetRequest;
import de.ozgcloud.zufi.grpc.fachstelle.GrpcFachstelleGetResponse;
import de.ozgcloud.zufi.grpc.fachstelle.GrpcFachstelleSearchRequest;
import de.ozgcloud.zufi.grpc.fachstelle.GrpcFachstelleSearchResponse;

class FachstelleRemoteServiceTest {

	@Spy
	@InjectMocks
	private FachstelleRemoteService service;

	@Mock
	private FachstelleServiceBlockingStub serviceStub;

	@Mock
	private FachstelleMapper fachstelleMapper;

	@Nested
	class TestSearch {

		private final String searchBy = LoremIpsum.getInstance().getWords(5);
		private final GrpcFachstelleSearchResponse grpcFachstelleSearchResponse = GrpcFachstelleSearchResponseTestFactory
				.create();

		@BeforeEach
		void setUpMocks() {
			when(serviceStub.search(argThat(requestContainsSearchBy()))).thenReturn(grpcFachstelleSearchResponse);
		}

		@Test
		void shouldCallGrpcStub() {
			callService();

			verify(serviceStub).search(argThat(requestContainsSearchBy()));
		}

		@Test
		void shouldCallMapper() {
			callService().toList();

			verify(fachstelleMapper).fromGrpc(GrpcFachstelleSearchResponseTestFactory.FACHSTELLE);
		}

		@Test
		void shouldReturnFachstelle() {
			var fachstelle = FachstelleTestFactory.create();
			when(fachstelleMapper.fromGrpc(GrpcFachstelleSearchResponseTestFactory.FACHSTELLE)).thenReturn(fachstelle);

			var fachstellen = callService();

			assertThat(fachstellen).containsExactly(fachstelle);
		}

		private Stream<Fachstelle> callService() {
			return service.search(searchBy);
		}

		private ArgumentMatcher<GrpcFachstelleSearchRequest> requestContainsSearchBy() {
			return request -> request.getSearchBy().equals(searchBy);
		}
	}

	@Nested
	class TestGetById {

		private final GrpcFachstelleGetRequest request = GrpcFachstelleGetRequestTestFactory.create();
		private final GrpcFachstelleGetResponse response = GrpcFachstelleGetResponseTestFactory.create();
		private final Fachstelle fachstelle = FachstelleTestFactory.create();

		@BeforeEach
		void setUp() {
			doReturn(request).when(service).buildGetByIdRequest(FachstelleTestFactory.ID);
			when(serviceStub.getById(request)).thenReturn(response);
			doReturn(fachstelle).when(service).getFachstelleFromGetByIdResponse(response);
		}

		@Test
		void shouldBuildRequest() {
			callService();

			verify(service).buildGetByIdRequest(FachstelleTestFactory.ID);
		}

		@Test
		void shouldCallGrpcStub() {
			callService();

			verify(serviceStub).getById(request);
		}

		@Test
		void shouldBuildResponse() {
			callService();

			verify(service).getFachstelleFromGetByIdResponse(response);
		}

		@Test
		void shouldReturnFachstelle() {
			var get = service.getById(FachstelleTestFactory.ID);

			assertThat(get).isEqualTo(fachstelle);
		}

		private Fachstelle callService() {
			return service.getById(FachstelleTestFactory.ID);
		}
	}

	@Nested
	class TestBuildGetByIdRequest {

		@Test
		void shouldHaveId() {
			var request = service.buildGetByIdRequest(FachstelleTestFactory.ID);

			assertThat(request.getId()).isEqualTo(FachstelleTestFactory.ID);
		}
	}

	@Nested
	class TestGetFachstelleFromGetByIdResponse {

		private final GrpcFachstelleGetResponse response = GrpcFachstelleGetResponseTestFactory.create();
		private final Fachstelle fachstelle = FachstelleTestFactory.create();

		@BeforeEach
		void setUp() {
			when(fachstelleMapper.fromGrpc(response.getFachstelle())).thenReturn(fachstelle);
		}

		@Test
		void shouldCallMapper() {
			callService();

			verify(fachstelleMapper).fromGrpc(response.getFachstelle());
		}

		@Test
		void shouldReturnFachstelle() {
			var built = callService();

			assertThat(built).isEqualTo(fachstelle);
		}

		private Fachstelle callService() {
			return service.getFachstelleFromGetByIdResponse(response);
		}
	}
}
