/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.historie;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.loeschanforderung.DeleteLoeschAnforderung;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;

class HistorieServiceTest {

	@Spy
	@InjectMocks
	private HistorieService service;
	@Mock
	private CommandService commandService;
	@Mock
	private HistorieCommandHandler historieCommandHandler;
	@Spy
	private List<HistorieProcessor> processors = new ArrayList<>();
	@Mock
	private HistorieProcessor processor;

	@DisplayName("Find finished commands")
	@Nested
	class TestFindFinishedCommands {

		@BeforeEach
		void mockProcessor() {
			processors.add(processor);
		}

		@DisplayName("process flow")
		@Nested
		class TestProcessFlow {

			private Command responseCommand = CommandTestFactory.create();

			@Captor
			private ArgumentCaptor<List<Command>> commandCaptor;

			@BeforeEach
			void initMock() {
				when(commandService.findFinishedCommands(any())).thenReturn(List.of(responseCommand).stream());
				when(historieCommandHandler.isHistorieCommand(any())).thenReturn(true);
				when(historieCommandHandler.translateOrder(any())).thenAnswer(i -> i.getArgument(0));

			}

			@Test
			void shouldCallService() {
				service.findFinishedCommands(CommandTestFactory.VORGANG_ID).toList();

				verify(commandService).findFinishedCommands(VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldFilterInvalidCommands() {
				service.findFinishedCommands(CommandTestFactory.VORGANG_ID).toList();

				verify(historieCommandHandler).isHistorieCommand(responseCommand);
			}

			@Test
			void shouldTranslateCommandOrderByCommandHandler() {
				service.findFinishedCommands(CommandTestFactory.VORGANG_ID).toList();

				verify(historieCommandHandler).translateOrder(responseCommand);
			}

			@Test
			void shouldProcessCommands() {
				service.findFinishedCommands(CommandTestFactory.VORGANG_ID).toList();

				verify(processor).process(commandCaptor.capture());
				assertThat(commandCaptor.getValue()).containsExactly(responseCommand);
			}

			@Test
			void shouldReturnProcessedCommands() {
				var processedCommands = Collections.singletonList(CommandTestFactory.create());
				when(processor.process(any())).thenReturn(processedCommands);

				var commands = service.findFinishedCommands(CommandTestFactory.VORGANG_ID).toList();

				assertThat(commands).isEqualTo(processedCommands);
			}
		}

		@DisplayName("Filter deleteLoeschAnforderung body commands")
		@Nested
		class TestFilterDeleteLoeschAnforderungBodyCommands {

			private final String deleteAttachedItemCommandId = UUID.randomUUID().toString();
			private final String changeStatusCommandId = UUID.randomUUID().toString();
			private final Map<String, Object> bodyMap = Map.of(DeleteLoeschAnforderung.DELETE_ATTACHED_ITEM_COMMAND_ID_FIELD,
					deleteAttachedItemCommandId,
					DeleteLoeschAnforderung.CHANGE_STATUS_COMMAND_ID_FIELD, changeStatusCommandId);
			private final Command deleteLoeschAnforderungCommand = CommandTestFactory.createBuilder()
					.order(CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN.name())
					.body(bodyMap).build();
			private final Command deleteAttachedItemCommand = CommandTestFactory.createBuilder().id(deleteAttachedItemCommandId).build();
			private final Command changeStatusCommandICommand = CommandTestFactory.createBuilder().id(changeStatusCommandId).build();

			private final List<Command> finishedCommands = List.of(deleteLoeschAnforderungCommand, deleteAttachedItemCommand,
					changeStatusCommandICommand);

			@BeforeEach
			void mock() {
				when(historieCommandHandler.isHistorieCommand(any())).thenReturn(true);
				when(commandService.findFinishedCommands(any())).thenReturn(finishedCommands.stream());
				when(historieCommandHandler.translateOrder(any())).thenAnswer(i -> i.getArgument(0));
				when(processor.process(any())).thenAnswer(i -> i.getArgument(0));
			}

			@Test
			void shouldFilterCommandsRelatedInBody() {
				var commands = service.findFinishedCommands(CommandTestFactory.VORGANG_ID).toList();

				assertThat(commands).hasSize(1).containsExactly(deleteLoeschAnforderungCommand);
			}
		}
	}
}