/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang.forwarding;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.InjectMocks;
import org.mockito.Spy;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.forwarding.Forwarding.Status;

@Component
class ForwardingModelAssemblerTest {

	@Spy // NOSONAR
	@InjectMocks
	private ForwardingModelAssembler modelAssembler;

	@Nested
	class TestToModel {

		@Test
		void shouldReturnEntityModel() {
			var model = modelAssembler.toModel(ForwardingTestFactory.create());

			assertThat(model).isNotNull();
		}

		@Test
		void shouldHaveSelfLink() {
			var model = modelAssembler.toModel(ForwardingTestFactory.create());

			assertThat(model.getLink(IanaLinkRelations.SELF)).isPresent().get()
					.extracting(Link::getHref).isEqualTo("/api/forwardings/" + ForwardingTestFactory.ID);
		}
	}

	@Nested
	class TestToCollectionModel {
		@Test
		void shouldCallToModel() {
			modelAssembler.toCollectionModel(Stream.of(ForwardingTestFactory.create()));

			verify(modelAssembler).toModel(any());
		}

		@Test
		void shouldHaveSelfLink() {
			var model = modelAssembler.toCollectionModel(Stream.of(ForwardingTestFactory.create()));

			assertThat(model.getLink(IanaLinkRelations.SELF)).isPresent().get()
					.extracting(Link::getHref).isEqualTo("/api/forwardings");
		}
	}

	@Nested
	class TestLinkOnModel {

		private final String expectedLinkHref = "/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/relations/" + ForwardingTestFactory.ID
				+ "/-1/commands";

		@Nested
		class TestMarkAsSuccessLink {

			@Test
			void shouldBePresentInStatusSENT() {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(Status.SENT).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_MARK_AS_SUCCESS)).isPresent().get()
						.extracting(Link::getHref).isEqualTo(expectedLinkHref);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.EXCLUDE, names = { "SENT" })
			void shouldNOTBePresentInAnyOtherStatus(Status status) {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(status).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_MARK_AS_SUCCESS)).isEmpty();
			}
		}

		@Nested
		class TestMarkAsFailLink {

			@ParameterizedTest
			@EnumSource(names = { "SENT", "SUCCESSFULL" })
			void shouldBePresentInStatusSENT(Status status) {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(status).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_MARK_AS_FAIL)).isPresent().get()
						.extracting(Link::getHref).isEqualTo(expectedLinkHref);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.EXCLUDE, names = { "SENT", "SUCCESSFULL" })
			void shouldNOTBePresentInAnyOtherStatus(Status status) {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(status).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_MARK_AS_FAIL)).isEmpty();
			}
		}

		@Nested
		class TestFailedLink {

			@Test
			void shouldBePresentInStatusFAILED() {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(Status.FAILED).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_FAILED)).isPresent().get()
						.extracting(Link::getHref).isEqualTo("/api/forwardings/" + ForwardingTestFactory.ID);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.EXCLUDE, names = { "FAILED" })
			void shouldNOTBePresentInAnyOtherStatus(Status status) {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(status).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_FAILED)).isEmpty();
			}
		}

		@Nested
		class TestSuccessfullLink {

			@Test
			void shouldBePresentInStatusSUCCESSFULL() {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(Status.SUCCESSFULL).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_SUCCESSFULL)).isPresent().get()
						.extracting(Link::getHref).isEqualTo("/api/forwardings/" + ForwardingTestFactory.ID);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.EXCLUDE, names = { "SUCCESSFULL" })
			void shouldNOTBePresentInAnyOtherStatus(Status status) {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(status).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_SUCCESSFULL)).isEmpty();
			}
		}

		@Nested
		class TestSendErrorLink {

			@Test
			void shouldBePresentInStatusSEND_ERROR() {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(Status.SEND_ERROR).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_ERROR)).isPresent().get()
						.extracting(Link::getHref).isEqualTo("/api/forwardings/" + ForwardingTestFactory.ID);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.EXCLUDE, names = { "SEND_ERROR" })
			void shouldNOTBePresentInAnyOtherStatus(Status status) {
				var model = modelAssembler.toModel(ForwardingTestFactory.createBuilder().status(status).build());

				assertThat(model.getLink(ForwardingModelAssembler.REL_ERROR)).isEmpty();
			}
		}
	}
}